;-----------------------------------------------------------
	global _decode_ax25_address
	global _encode_ax25_address

	global _addr_equals
	global _addr_starts

	global _tidy_callsign

	global _axacpy
	global _axacmp
	global _axacmp_prefix


;===========================================================
	section .text2 ; XXX not fncallable
	align 1

;-----------------------------------------------------------
; contortions: show embedded spaces in callsigns (diagnostics)

_decode_ax25_address  ; (char *call_buffer, const byte *addr)

	xgey
	xgdx

	ldaa 6,x       ; save ssid byte while the pointer is at [0]
	lde  # 5       ; 5,x last callsign byte
L40
	ldab e,x       ; pick from right
	lsrb
	cmpb # ' '
	bne  L41       ; found the last significant. now sprint.
	sube # 1
	bne  L40       ; while not at addr[0]
L41
	ldab 0,x       ; pick from left
	aix  # 1
	lsrb
	stab 0,y       ; always at least one character sprinted.
	aiy  # 1
	sube # 1
	bcc  L41       ; while significant characters left

	anda # 30      ; 15 << 1, ssid mask
	beq  L43
	lsra

	ldab # '-'
	stab 0,y
	aiy  # 1

	cmpa # 10
	blo  L42
	suba # 10

	ldab # '1'
	stab 0,y
	aiy  # 1
L42
	adda # '0'
	staa 0,y
	aiy  # 1
L43
	clr  0,y
	xgdy          ; return pointer to \0 in dst buffer
	rts

;-----------------------------------------------------------
; contortions: 'call' can end in whitespace or punctuation, or \0

_encode_ax25_address  ; byte * (byte *abuf, const char *call)

	xgey
	xgdx
	clre
L20
	ldab 0,x
	aix  #+1
	tba

	subb # 'A'
	cmpb # 'Z'-'A'
	bls  L22        ; upper
	addb # 'A'-'0'
	cmpb # 9
	bls  L22        ; digit
	subb # 'a'-'0'
	cmpb # 'z'-'a'
	bls  L21        ; lower
	aix  #-1        ; not alnum anymore. stay here.
	clra
L21
	eora # 20h      ; lower to upper, or \0 to space
L22
	asla
	staa e,y
	adde # 1
	cpe  # 6
	blo  L20

	clrb
	ldaa 0,x
	cmpa # '-'
	bne  L23
	bsr  get_ssid
L23
	aslb
	orab # 60h      ; reserved bits
	stab 6,y
	xgdy
	addd # 7        ; return &abuf[7], ready for next slot.
	rts

;-----------------------------------------------------------
get_ssid
	clrb
	ldaa 1,x       ; first, digit 0...9 ?
	suba # '0'
	cmpa # 9
	bhi  L0        ; no ssid
	tab
	cmpa # 1
	bne  L0        ; ssid 0 or 2...9

	ldaa 2,x       ; second. digit 0...5 ?
	suba # '0'
	cmpa # 5
	bhi  L0        ; just -1 it seems.
	tab
	addb # 10      ; -10 to -15
L0
	rts

;-----------------------------------------------------------
_addr_equals       ; bool (byte *a, char *call)

	xgey
	xgdx
	clre
L10
	ldaa e,y       ; from 'a'
	lsra
	cmpa # ' '
	beq  L11       ; end of callsign part of 'a'
	cmpa e,x
	bne  false     ; callsign part mismatch
	adde # 1
	cpe  # 6
	blo  L10
L11
	ldab e,x
	beq  L12          ; 'call' ended too, verify ssid 0
	cmpb # '-'
	bne  false        ; not "" nor "-*",  reject

	aex
	bsr  get_ssid
L12
	aslb
	eorb 6,y
	andb # 30      ; 15 << 1
	bne  false
true
	ldd  # 1         ; TRUE
	rts

;-----------------------------------------------------------
_addr_starts      ; bool (byte *a, char *prefix)

	xgey	
	xgdx	
	lde  # -1
L30
	adde # 1
	ldab e,x
	beq  true     ; prefix ends before mismatch
	aslb
	cmpb e,y
	beq  L30      ; while match
false
	clrd
	rts

;-----------------------------------------------------------
_tidy_callsign
	pshm z
	ais  #-4
	tsz

	ste  0,z
	jsr  _strend
	xgdy
L50
	cpy  0,z
	beq  L59         ; empty string (now)
	sty  2,z
	ldab -1,y
	clra
	tde
	jsr  _isalnum
	ldy  2,z
	tstb
	bne  L51         ; alnum, to the left

	aiy  #-1
	clr  0,y         ; snip
	bra  L50         ; more spaces ?
L51
	aiy  #-2
	cpy  0,z
	blo  L52
	ldd  0,y         ; last two characters
	cpd  # '-0'
	bne  L52
	clr  0,y         ; snip snip
L52
	lde  0,z
	jsr  _strupper
L59
	ais  #+4
	pulm z
	rts

;-----------------------------------------------------------
_axacpy
	xgex           ; copy to
	xgdy           ; copy from
	lde  # 5       ; copy callsign
axacpy_1
	ldab e,y
	stab e,x       ; no bits are masked here
	sube # 1
	bcc  axacpy_1  ; } while (--i >= 0);

	ldab 6,y       ; copy SSID
	andb # 7Eh     ; but not cmd/resp/repeated nor last
	stab 6,x
	rts

;===========================================================
	section .inthand  ; nano-optimisation
	align 1

;-----------------------------------------------------------
_axacmp
	xgex           ; shifted ascii addr1
	xgdy           ; shifted ascii addr2
	lde   # 5      ; start from [5]
L67
	ldab  e,x
	eorb  e,y
	lsrb           ; unnecessary really
	bne   L68      ; different
	sube  # 1
	bcc   L67      ; more characters

	ldab  6,x
	eorb  6,y
	andb  # 1Eh    ; SSID bits
L68
	clra           ; 0 if equal
	rts

;-----------------------------------------------------------
_axacmp_prefix
	xgex           ; address to compare
	xgdy           ; prefix to compare against
	clre           ; start from [0]
pfx_1
	ldab  e,y
	beq   pfx_out  ; end of prefix
	eorb  e,x
	adde  # 1      ; i++
	lsrb
	beq   pfx_1    ; while same
pfx_out
	clra
	rts            ; 0 if equal prefixes

;===========================================================
	section .text2  ; back to normal XXX not fncallable
	align 1

;-----------------------------------------------------------
