
/*
 *  Autodetect make:
 *  $PGRM        Garmin
 *  $PMGN        Magellan
 *  $PMVX        Magnavox ? Philips ? Leica ?
 *  $PHMS        ditto
 *  $PMOT        Motorola
 *  $PRWI        Rockwell
 *  $PSRF        Sirf
 *  $PSNY        Sony
 *  $PRAY        Raytheon
 *  $PTNL        Trimble
 *  $PNTS        Navitora, fabricated
 */

#include "standard.h"
#include "time.h"
#include "hamdr_cfg.h"
#include "fix.h"

static void
mate(byte *astral_eartha)
{
	static byte suppress, echo; /* silly */

	if (suppress ^= 1) {
		SCI_puts("\r\n");
		SCI_puts(astral_eartha);
		SCI_puts("\r\n");
	}
}

/* SiRF -------------------------------------------------------------- */

/* motorola order suits us */

static void
sirf_word(word x)
{
	SCI_putc(x >> 8); /* MSByte */
	SCI_putc(x);
}

static void
SiRF_initialize()
{
	static byte msg[] = {
		0x81,       /* Message ID - Switch To NMEA */
		0x02,       /* Mode - 2 = NMEA debug neither enabled nor disabled */
		1, 1,       /* GGA interval and checksum flag - 1 Hz */
		 0, 1,      /* GGL interval and checksum flag */
		 0, 1,      /* GSA interval and checksum flag */
		 0, 1,      /* GSV interval and checksum flag */
		1, 1,       /* RMC interval and checksum flag - 1 Hz */
		 0, 1,      /* VTG interval and checksum flag */
		 0, 1,      /* MSS interval and checksum flag */
		 0, 1,      /*  future sentence */
		 0, 1,      /* ZDA interval and checksum flag */
		 0, 1,      /*  future sentence */
		0x12, 0xC0, /* Baud Rate - 4800 baud XXX see SCI_config below */
	};
	byte *p;

	SCI_config(38400, 0);

	/*
	 *  start sequence
	 *  payload length
	 *  payload
	 *  message checksum, MSbit clear, over payload only
	 *  end sequence
	 */
	sirf_word(0xA0A2);
	sirf_word(sizeof(msg));

	p = msg;
	do {
		SCI_putc(*p);
	} while (++p < __ENDOF(msg));

	sirf_word(checksum(msg, sizeof(msg)) & 0x7FFF);
	sirf_word(0xB0B3);


	SCI_drain();             /* make sure the bytes get out safely */
	SCI_config(4800, 0);     /* XXX see Baud Rate in msg above */
}

/* EarthMate et al Zodiac speakers ----------------------------------- */

#if 0

STATIC word rockwell_to_NMEA[] = {
	0x81FF,                  /* sync */
	1331,                    /* id */
	3,                       /* num_words */
	0,                       /* flags */
	- 0x81FF - 1331 - 3 - 0, /* hdr_cksum */

	0,                       /* seq_num */
	0,                       /* reserved */
	1,                       /* NMEA = 1 */
	- 0 - 0 - 1,             /* data_cksum */
};

/* intel order, flip */

static void
zodiac_word(word x)
{
	SCI_putc(x);      /* LSByte */
	SCI_putc(x >> 8);
}

#endif

/* only words and longs in structure, intel order */

STATIC struct {
	long set_time;             /* 6,7 UDI 10 msec ticks 0..4G */
	word seq_num;
	word sat_meas_seq_num;
	word solution_validity_bits; /* bit set = invalid, not all are fatal */
	word solution_type_bits;
	word solution_meas_count;    /* does this mean #sats utilized ? */
	word polar_navigation;
	word gps_week;
	long gps_seconds;
	long gps_nanoseconds;
	word utc_day;              /* 19    UI  day     1..31 */
	word utc_month;            /* 20    UI  month   1..12 */
	word utc_year;             /* 21    UI  year    1980..2079 */
	word utc_hours;            /* 22    UI  hours   0..23 */
	word utc_minutes;          /* 23    UI  minutes 0..59 */
	word utc_seconds;          /* 24    UI  seconds 0..59 */
	long utc_nanoseconds;
	long latitude;             /* 27,28 DI  radians +-pi/2  10^-8 */
	long longitude;            /* 29,30 DI  radians +-pi    10^-8 */
	long height;               /* 31,32 DI  meters  +-50000 10^-2 */
	word geoidal_separation;
	long ground_speed;         /* 34,35 UDI m/s     0..1000 10^-2 */
	word true_course;          /* 36    UI  radians 0..2pi  10^-3 */
	word magnetic_variation;
	word climb_rate;
	word map_datum;
	long hposerr;
	long vposerr;
	long timerr;
	word spderr;
	long clock_bias;
	long clock_bias_stddev;
	long clock_drift;
	long clock_drift_stddev;   /* 53,54 */
	word data_cksum;           /* 55    not included in num_words - oddity */
} msg1000;

static void
zodiac_cvt(long *px, long *py, long *scale)
{
	*py = ld_long_intel(px);
	SCALE32_signed(py, scale);
}

static void
zodiac_gps_update()
{
	static long
		coord_scale = 147650099,  /* 10800000 * 4G / 314159265 */
		alt_scale   = 42949673;   /* cm to m */

	long alt;
	struct tm tm;
	struct vector v;

	/* XXX is this proper ? */

	gps_satellites = ld_word_intel(&msg1000.solution_meas_count);

	tm.year   = ld_word_intel(&msg1000.utc_year) % 100;
	tm.month  = ld_word_intel(&msg1000.utc_month);
	tm.day    = ld_word_intel(&msg1000.utc_day);

	tm.hour   = ld_word_intel(&msg1000.utc_hours);
	tm.minute = ld_word_intel(&msg1000.utc_minutes);
	tm.second = ld_word_intel(&msg1000.utc_seconds);

	/* not enough satellites or exceeded maximum E?PE ? */

	if (ld_word_intel(&msg1000.solution_validity_bits) & 0x1C) {
		gps_nofix(&tm);
		return;
	}

	/* latitude/longitude -- signed radian / 1e8
	 * height -- signed centimeters */

	zodiac_cvt(&msg1000.latitude,  &v.coord.lat, &coord_scale);
	zodiac_cvt(&msg1000.longitude, &v.coord.lon, &coord_scale);

	zodiac_cvt(&msg1000.height, &alt, &alt_scale);

	/*
	 *  ground_speed -- unsigned m/s / 1e2
	 *  0.01 m/s to knots
	 *
	 *  true_course -- unsigned radian / 1e3
	 *  0.001 radians to degrees: *= 0.05729578 ~ 653/11397
	 */
	v.knots  = muldiv(ld_dword_intel(&msg1000.ground_speed), 3600 / 100, 1852);
	v.course = muldiv(ld_word_intel(&msg1000.true_course), 653, 11397);

	/* polar ? force abs(longitude)=90deg and course=south */

	sted_gps_altitude(alt);

	gps_fix(&v, &tm);
}

static bool
zodiac_cksum(void *p, int n)
{
	word *wp = p;
	word sum = 0;

	do {
		sum += ld_word_intel(wp++);
	} while (--n);

	return (sum == 0);
}

/*
 *  sync               0x81FF
 *  id                 1000 (else ignore message)
 *  num_words          55 - 5  - 1 = 49
 *  flags              dont care
 *  hdr_cksum          such that sum(hdr) == 0
 *  [data data_cksum]  num_words + 1 iff any words
 */
static void
zodiac_gps_input(word c)
{
	static word i, id, num_bytes;
	static byte hdr[5 * 2];         /* 5 words, 10 bytes in header */

	if (i < sizeof(hdr)) {
		hdr[i] = c;
		if (++i != sizeof(hdr))
			return;

		if (ld_word_intel(&hdr[0]) == 0x81FF
		 && zodiac_cksum(hdr, sizeof(hdr) / 2)) {

			word wc;

			id = ld_word_intel(&hdr[2]);
			wc = ld_word_intel(&hdr[4]);

			if (wc == 0)
				goto flush;      /* no associated data */

			num_bytes = (wc + 1) * 2;
			return;
		}
		/* resync until header is valid. should not happen often.
		 * also, catch "....EARTHA" and reply to it */

		if (!substrcmp(hdr + 4, "EARTHA")) {
			mate("EARTHA");
			goto flush;
		}
		bcopy(hdr + 1, hdr, --i);
		return;
	}

	/* now in data area. messages other than 1000 are discarded */

	if (id == 1000 && i < sizeof(hdr) + sizeof(msg1000))
		((byte *) &msg1000)[i - sizeof(hdr)] = c;

	if (++i < sizeof(hdr) + num_bytes)
		return;

	if (id == 1000 && zodiac_cksum(&msg1000, sizeof(msg1000) / 2))
		zodiac_gps_update();
flush:
	i = 0;
}

/* Aisin-Seiki ------------------------------------------------------- */

/* motorola order. the alignment does not cooperate though */


/* full, complete, auto-transmit sequence */

STATIC
struct {

	/* CACA 03
	   035AA267 018B34F5 0064 F600
	   03AA FF
	   0000 09
	   010A00
	   060321 154728
	   12 00
	   1C 01C8100B0500DEE0
	   00 B10D */

	struct { byte tag[2],
		validity,
		latitude[4], longitude[4], height[2], error_ellipse[2],
		heading[2], heading_error,
		speed[2], speed_error,
		pad1[3],
		datetime[6],         /* yymmddHHMMSS as 6 bcd bytes */
		HDOP, VDOP,
		satellite_count, satellite_info[8],
		pad2,
		checksum, terminator; } CA; /* 44 bytes */

	/* CBCB 28
	   1708239560
	   1408375F56
	   1908342C38
	   01120D2F1F
	   180100AC4C
	   040100D631
	   0D0100A52E
	   020100E50F 00FFF7FF1F D30D */

	struct { byte tag[2],
		datetime_second;
		struct aisin_sat_info { byte
			id,
			condition,
			level,
			azimuth,
			elevation;
		}
			satellite_information[8];
		byte
			almanac[5],
		checksum, terminator; } CB; /* 50 bytes */

	/* C3C3 02 0A 00 01 00 0000 6D0D */

	struct { byte tag[2],
		receiver_mode,
		elevation_mask,
		signal_level_mask,
		PDOP_switch,
		smooth_level,
		set_height[2],
		checksum, terminator; } C3; /* 11 bytes */

	/* C4C4 8587B0 2A 000000 920D */

	struct { byte tag[2],
		version_code[3],
		unsettled_value,
		diagnosis[3],
#define DIAG_ANT_OPEN  0x51
#define DIAG_ANT_SHORT 0x53
		checksum, terminator; } C4; /* 11 bytes */

} aisin;               /* 44 + 50 + 11 + 11 == 116 bytes */


/* unit of *px is 1/256 of a second. signed */

static void
aisin_coord(long *px, long *mm)
{
	static long scale = 279620266; /* second/256 to milliminute */

	*mm = *px;
	SCALE32_signed(mm, &scale);
}

static void
aisin_gps_update()
{
	struct tm tm;
	struct vector v;
	word nsat;

	if (aisin.CA.validity == 0xFF && cfg.lcd) {

		/* just a generic "ERROR", unless another of these less grave,
		 * but common, errors is the worst diagnostic */

		lcd_row(1);
		switch (aisin.C4.diagnosis[0]) {
		case DIAG_ANT_SHORT: lcd_str("GPS ANT SHORTED"); break;
		case DIAG_ANT_OPEN:  lcd_str("GPS ANT OPEN");    break;
		default:             lcd_str("GPS ERROR");       break;
		}
		lcd_finis();
	}
	/* unpack date and time from BCD bytes to binary - bit nasty */

	repack_bcd_to_binary(aisin.CA.datetime, 6);

	tm.year   = aisin.CA.datetime[0];
	tm.month  = aisin.CA.datetime[1];
	tm.day    = aisin.CA.datetime[2];

	tm.hour   = aisin.CA.datetime[3];
	tm.minute = aisin.CA.datetime[4];
	tm.second = aisin.CA.datetime[5];

	/* DOP unit 0.2 */

	HDOP = aisin.CA.HDOP / 5;
	VDOP = aisin.CA.VDOP / 5;

	nsat = aisin.CA.validity & 0x0F;
	switch (nsat) {
	default:
		nsat = 0;
	case 0:
	case 1:
	case 2:
		gps_satellites = nsat;
		gps_nofix(&tm);
		return;                  /* no fix, bail out */

	/* 2D or 3D fix */
	case 3:
	case 4: {
			struct aisin_sat_info *p = aisin.CB.satellite_information;
			nsat = 0;
			do {
				if (p->condition & 0x0C) /* decoded */
					++nsat;
			} while (++p < __ENDOF(aisin.CB.satellite_information));
		}
	}

	/* lat/lon --- unit is 1/256 of a second */

	aisin_coord((long *) aisin.CA.latitude,  &v.coord.lat);
	aisin_coord((long *) aisin.CA.longitude, &v.coord.lon);

	/*
	 * speed --- 1/4 m/s to knots
	 * speed range is 0...333 or 83.3 m/s or 300 km/h or 162 knots.
	 * speed out of range is encoded as FFFF
	 * heading, 1024 steps of 360 degrees.
	 * height, unit 0.5 m, range -100m ... 6000m
	 */
	v.knots = *(word *) aisin.CA.speed;
	if (v.knots < 333)
		v.knots = muldiv(v.knots, 3600 / 4, 1852);
	else
		v.knots = 162; /* needle at max, 300 km/h */

	v.course = muldiv(*(word *) aisin.CA.heading, 360, 1024);


	sted_gps_altitude(*(short *) aisin.CA.height >> 1);

	gps_satellites = nsat;

	gps_fix(&v, &tm);
}

/*
 * Data is TTL level 9600 baud 8 bits _plus_ even parity
 */
static void
aisin_gps_input(word c)
{
	static word i;
	byte *sync;
#define aisin_caddr ((byte *) &aisin)

	if (i < 2 && c != 0xCA)
		goto flush;

	aisin_caddr[i++] = c;

	if (i < sizeof(aisin))
		return;

	/* checksum does not include the 0x0D terminator, thus -1 */

	if (*(word *) aisin.CA.tag == 0xCACA && aisin.CA.terminator == 0x0D
	 && *(word *) aisin.CB.tag == 0xCBCB && aisin.CB.terminator == 0x0D
	 && *(word *) aisin.C3.tag == 0xC3C3 && aisin.C3.terminator == 0x0D
	 && *(word *) aisin.C4.tag == 0xC4C4 && aisin.C4.terminator == 0x0D
	 && (byte) checksum(&aisin.CA, sizeof(aisin.CA) -1) == 0
	 && (byte) checksum(&aisin.CB, sizeof(aisin.CB) -1) == 0
	 && (byte) checksum(&aisin.C3, sizeof(aisin.C3) -1) == 0
	 && (byte) checksum(&aisin.C4, sizeof(aisin.C4) -1) == 0) {

		aisin_gps_update();
		goto flush;
	}
	sync = memchr(aisin_caddr + 2, 0xCA, sizeof(aisin) - 2);
	if (sync) {
		i = aisin_caddr + sizeof(aisin) - sync;
		bcopy(sync, &aisin, i);
		return;
	}
flush:
	i = 0;
}

/* SeaTalk ----------------------------------------------------------- */

static struct {
	struct seatalk_coord {
		byte SW;
		byte deg;
		word mm;
	} lat, lon;
	word course, knots;  /* integer */
	struct tm tm;
} seatalk;

static void
seatalk_coord(struct seatalk_coord *ord, long *pmm)
{
	*pmm = EMUL(_DEG_AS_mm, ord->deg) + ord->mm;
	if (ord->SW)
		lnegate(pmm);
}

static void
seatalk_gps_update()
{
	struct vector v;

	seatalk_coord(&seatalk.lat, &v.coord.lat);
	seatalk_coord(&seatalk.lon, &v.coord.lon);

	v.course = seatalk.course;
	v.knots  = seatalk.knots;

	gps_fix(&v, &seatalk.tm);
}

/*
 * 53 U0 VW      (U & 3) * 90 + (VW & 0x3F) * 2 + (U & 0xC) / 8 = course
 */
static word
seatalk_course(word U0, word VW)
{
	return ((U0 & 0x30) >> 3) * 45 + (VW & 0x3F) * 2; /* XXX huh ? */
}

static void
seatalk_message(byte *msg)
{
	word x;

#if DEBUG
	if (debug) {
		print_hexdump(msg, (msg[1] & 15) + 3);
		eol();
	}
#endif
	switch (msg[0]) {

	/*
	 * 50 A2 XX YYYY latitude,  XX deg, YYYY / 100 = min YYYY.15 = South
	 * 51 A2 XX YYYY longitude, XX deg, YYYY / 100 = min YYYY.15 = East
	 * 58 Z5 LA XXYY LO QQRR
	 *                 LA and LO deg, XXYY / 1000 and QQRR / 1000 = min
	 *                 Z.0 = South, Z.1 = East
	 */
	case 0x50:
		x = ld_word_intel(&msg[3]);
		seatalk.lat.deg = msg[2];
		seatalk.lat.mm  = (x & 0x7FFF) * 10;
		seatalk.lat.SW  = (x & 0x8000) ? TRUE : FALSE;
		break;

	case 0x51:
		x = ld_word_intel(&msg[3]);
		seatalk.lon.deg = msg[2];
		seatalk.lon.mm  = (x & 0x7FFF) * 10;
		seatalk.lon.SW  = (x & 0x8000) ? FALSE : TRUE;
		break;

	case 0x58:
		seatalk.lat.deg = msg[2];
		seatalk.lat.mm  = ld_word_intel(&msg[3]);
		seatalk.lat.SW  = (msg[1] & 1) ? TRUE : FALSE;

		seatalk.lon.deg = msg[5];
		seatalk.lon.mm  = ld_word_intel(&msg[6]);
		seatalk.lon.SW  = (msg[1] & 2) ? FALSE : TRUE;
		break;

	/*
	 * 52 01 XXXX    XXXX / 10 = knots
	 * 53 U0 VW      (U & 3) * 90 + (VW & 0x3F) * 2 + (U & 0xC) / 8 = course
	 */
	case 0x52:
		seatalk.knots = ld_word_intel(&msg[2]) / 10;
		break;

	case 0x53:
		seatalk.course = seatalk_course(msg[1], msg[2]); /* cc1.core */
		break;

	/*
	 * 54 t1 TT HH   HH hours, (TTt & 0xFC0) / 64 = minutes, TTt & 0x3F = secs
	 * 56 M1 DD YY   YY year, M month, DD day in month
	 */
	case 0x54:
		x = ld_word_intel(&msg[1]) >> 4;
		seatalk.tm.hour   = msg[3];
		seatalk.tm.minute = msg[2] >> 2;
		seatalk.tm.second = x & 0x3F;

		seatalk_gps_update(); /* XXX which msg ends "current solution" ? */
		break;

	case 0x56:
		seatalk.tm.year  = msg[3];
		seatalk.tm.month = msg[1] >> 4;
		seatalk.tm.day   = msg[2];
		break;
#if 0
	/*
	 * 6E 07 00 00 00 00 00 00 00 00  MOB
	 * 36 00 01                       Cancel MOB condition
	 * A5 57 QQ HH ?? AA GGZZ YY DD   GPS and DGPS fix info
	 */
	case 0xA5:
		/* gps_nofix(NULL); ??? XXX ??? */
		break;
#endif
	}
}

static void
seatalk_gps_input(word c) /* XXX */
{
	static byte buf[3 + 15];
	static word i = sizeof(buf);  /* discard until fresh message */

	if (c > 0xFF)
		i = 0;

	if (i >= sizeof(buf))
		return;

	buf[i++] = c;

	if (i >= 3 && (buf[1] & 15) + 3 == i)
		seatalk_message(buf);
}

/* Oncore ------------------------------------------------------------ */

#if 0

static void
oncore_send(byte *p, word cc)
{
	byte cksum = 0;

	SCI_putc(0x40);
	SCI_putc(0x40);

	do {
		cksum ^= *p;
		SCI_putc(*p);
		++p;
	} while (--cc);

	SCI_putc(cksum);

	SCI_putc(0x0D); /* CR LF */
	SCI_putc(0x0A);
}

static void
oncore_init()
{
	SCI_drain();
	SCI_config(9600, 0);

	oncore_send("Cg\001", 3); /* terminate idle mode */
	oncore_send("Cg\001", 3); /* terminate idle mode */

	oncore_send("Ha\001", 3); /* 12 CH message every second */
	oncore_send("Hb\001", 3); /* 12 CH short position message every second */
	oncore_send("Ea\001", 3); /*  8 CH message every second */
	oncore_send("Ba\001", 3); /*  6 CH message every second */

	oncore_send("Ci\001", 3); /* switch to NMEA */
}
#endif

/*
 * @@Bamdyyhmsffffaaaaoooohhhhmmmmvvhhddtnt imsd*6 sC\r\n 68 bytes
 * @@Eamdyyhmsffffaaaaoooohhhhmmmmvvhhddtnt imsd*8 sC\r\n 76 bytes
 *
 * s at end, the receiver status byte:
 *		Bit 7: Position Propagate mode
 *		Bit 6: Poor Geometry (DOP over 20)
 *		Bit 5: 3D fix
 *		Bit 4: Altitude Hold (2D fix)
 *		Bit 3: Acquiring Satellites/Position Hold
 *		Bit 2: Differential
 *		Bit 1: Insufficient visible satellites (less than 3)
 *		Bit 0: Bad Almanac
 *
 * @@HbmdyyhmsffffaaaaoooohhhhmmmmVVvvhhddnt ssrrvvvvvvC\r\n 54 bytes
 * @@HamdyyhmsffffaaaaoooohhhhmmmmaaaaoooohhhhmmmmVVvvhhddnt imsidd*12
 *       ssrrccooooTTushmvvvvvvC\r\n 154 bytes
 *
 * ss, the receiver status word
 *		Bit 15-13: 111 = 3D Fix
 *				   110 = 2D Fix
 *				   101 = Propagate Mode
 *				   100 = Position Hold
 *				   011 = Acquiring Satellites
 *				   010 = Bad Geometry
 *				   001 = Reserved, 000 = Reserved
 *		Bit 12-11: Reserved
 *		Bit 10: Narrow band tracking mode (timing rx only)
 *		Bit 9: Fast Acquisition Position
 *		Bit 8: Filter Reset To Raw GPS Solution
 *		Bit 7: Cold Start
 *		Bit 6: Differential Fix
 *		Bit 5: Position Lock
 *		Bit 4: Autosurvey Mode
 *		Bit 3: Insufficient Visible Satellites
 *		Bit 2-1: Antenna Sense 00 = OK
 *							   01 = OC
 *							   10 = UC
 *							   11 = NV
 *		Bit 0: Code Location 0 = EXTERNAL, 1 = INTERNAL
 */

struct oncore_HaHbEaBa_head {
	word
		id;             /* message identifier */
	byte
		month,          /* 1 .. 12             Date */
		day;            /* 1 .. 31 */
	word
		year;           /* 1980 .. 2079 */
	byte
		hours,          /* 0 .. 23             Time */
		minutes,        /* 0 .. 59 */
		seconds,        /* 0 .. 60 */
		frac_sec   [4], /* 0 .. 999,999,999 - alignment lost, thus [4] */

		latitude   [4], /* signed milliarcseconds (+-90 deg) */
		longitude  [4], /* (+-180 deg) */

		height_ref [4], /* -100,000 .. 1,800,000 cm */
		height_MSL [4]; /* -100,000 .. 1,800,000 cm */
};

struct oncore_vvhh {
	word
		velocity,  /* 0 .. 51400 cm/s (0 to 514.00 m/s) */
		heading;   /* 0 .. 3599 0.1 deg (0.0 to 359.9 o) */
};

static void
oncore_cvt(long *px, long *py, word div)
{
	*py = EDIVx(labs(*px), div, NULL);
	if (*px < 0)
		lnegate(py);
}

static void
oncore_gps_update(struct oncore_HaHbEaBa_head *p, struct oncore_vvhh *vh)
{
	long alt;
	struct tm tm;
	struct vector v;

	tm.year   = p->year % 100;
	tm.month  = p->month;
	tm.day    = p->day;

	tm.hour   = p->hours;
	tm.minute = p->minutes;
	tm.second = p->seconds;

	/* milliarcseconds to milliminutes,
	 * centimeters to meters. */

	oncore_cvt((long *) p->latitude,  &v.coord.lat, 60);
	oncore_cvt((long *) p->longitude, &v.coord.lon, 60);

	oncore_cvt((long *) p->height_MSL, &alt, 100);

	/* cm/s to knots and 0.1 degrees to integer degrees */

	v.knots  = muldiv(vh->velocity, 3600 / 100, 1852);
	v.course = vh->heading / 10;

	sted_gps_altitude(alt);

	gps_fix(&v, &tm);
}

static void
oncore_gps_message(byte *msg, word len)
{
	byte *vvhh = msg + 29;
	bool valid;

	switch (*(word *) msg) {
	default:
		return;

	/* accept Propagate Mode, 2D or 3D */

	case 0x4261:                       /* Ba */
		valid = (msg[62] & 0xB0) != 0;
		break;

	case 0x4561:                       /* Ea */
		valid = (msg[70] & 0xB0) != 0;
		break;

	case 0x4862:                       /* Hb */
		vvhh = msg + 31;
		valid = (msg[39] & 0xE0) >= 0xA0;
		break;

	case 0x4861:                       /* Ha */
		vvhh = msg + 47;
		valid = (msg[127] & 0xE0) >= 0xA0;
		break;
	}
	if (valid)
		oncore_gps_update((void *) msg, (void *) vvhh);

#if 0 /* let it timeout */
	else
		gps_nofix(NULL);
#endif
}

static void
oncore_gps_input(word c)
{
	static byte sum, len, buf[256];

	sum ^= c;
	buf[len++] = c;   /* wraps to empty if no sync after 256 bytes */

	if (len < 4)
		return;

	if (len == 4) {
		if (*(word *) buf != 0x4040
		 || (buf[2] < 'A' || buf[2] > 'Z')
		 || (buf[3] < 'a' || buf[3] > 'z')) {
			buf[0] = buf[1];
			buf[1] = buf[2];
			buf[2] = buf[3];
			len = 3;
			return;
		}
		sum = buf[2] ^ buf[3];
		return;
	}
	if (len > 7 && sum == 0x0D ^ 0x0A && *(word *) &buf[len - 2] == 0x0D0A) {

		oncore_gps_message(buf + 2, len - 5);
		len = 0;
	}
}

/* common ------------------------------------------------------------ */

/* cfg.gps == 0 bypasses this */

void
quirky_gps_input(word c)
{
	switch (cfg.gps) {

	case GPS_GENERIC:      /* unnecessary, but save a microsecond or two */
	case GPS_GARMIN:
	case GPS_MAGELLAN:
	case GPS_SIRF:
	case GPS_TRIPMATE:
	default:
		nmea_gps_input(c);
		break;

	case GPS_EARTHMATE:
	case GPS_ZODIAC:
		zodiac_gps_input(c);
		break;

	case GPS_AISIN:
		aisin_gps_input(c);
		break;

	case GPS_SEATALK:
		seatalk_gps_input(c);
		break;

	case GPS_ONCORE:
		oncore_gps_input(c);
		break;

	case GPS_TSIP:
		tsip_gps_input(c);
		break;

#if 0
	case GPS_AUTO:             /* unknown, yet */
		aisin_gps_input(c);
		nmea_gps_input(c);
		oncore_gps_input(c);
		seatalk_gps_input(c);
		tsip_gps_input(c);
		zodiac_gps_input(c);
		break;
#endif
	}
}

/* command interpreter did not know about 'cmd'.
 * maybe it is one of the odd gps's ? speed etc ok by now */

int
gps_babble_maybe(byte *cmd)
{
	if (!strcmp(cmd, "ASTRAL")) {
		mate("ASTRAL");            /* normal NMEA from now on */

		cfg.gps = GPS_TRIPMATE;
		return (TRUE);
	}
	if (!strcmp(cmd, "EARTHA")) {
		mate("EARTHA");            /* rockwell/zodiac from now on */

		cfg.gps = GPS_EARTHMATE;
		return (TRUE);
	}
	return (FALSE);
}

void
gps_initialize()
{
	/* allow any messages to get thru */

	SCI_drain();

	if (cfg.ttybaud == 0)
		SCI_config(4800, 0);     /* unless overridden in cfg */

	switch (cfg.gps) {

	case GPS_AISIN:              /* proprietary format */
		SCI_config(9600, 2);     /* Aisin must be 9600 8E1 */
		break;

	case GPS_SIRF:               /* command it to emit NMEA */
		SiRF_initialize();
		break;

	case GPS_TRIPMATE:           /* encourage it to emit NMEA */
		SCI_config(4800, 0);
		break;

	case GPS_EARTHMATE:          /* uses rockwell zodiac format */
	case GPS_ZODIAC:             /* proprietary format XXX some do NMEA ? */
	case GPS_ONCORE:          /* assume these are those not capable of NMEA */
		SCI_config(9600, 0);
		break;

	case GPS_SEATALK:
		SCI_config(4800, 3);     /* 9 bit characters */
		break;
	}
}

/* Sony IPS et al ---------------------------------------------------- */

#if 0

Sony IPS-2k/3k/5k (?) format, they say, is like this, 110 bytes long.
Looks like OCR exercise, suspect examples.

SONY 80
950701 6 090346 N3546569 E13918458 +0218 004 013
950701 6 090345 D 4 B
DHIFG XHbCI RDFFF PEiFH SCKCQ GBRFF eBEDD cCOCH
d DH 1 0
\r\n

Another strange format is ETAK (?), 130 bytes long, just like sony_ips,
but there is 20 additional bytes between 'datum' and 'channel_info'.
Looks like it just has more channels.

SMAT C0
980927 0 183213 N3542431 E13940033 +0051 000 265
980927 0 183213 Q 3 B
BEeDA CAKDA NEPF\ OHCF[
PBqDA SBoDA cCkDA eAHDA gDJFY ___AA ___AA ___AA
A DB E E
\r\n

some fields are ________ until first fix is calculated ?
weekday is letter until time valid ?

    SONY : fixed tag
      80 : version, lots of different values
  950701 : current date/time, yymmdd weekday hhmmss, UTC
       6   0=Sunday
  090346
       N : North or South latitude, n/s if not current (fix?)
 3546569 : 35 deg 46.569 minute _OR_ 35 deg 46 min 56.9 second, see below.
       E : East or West longitude
13918458
   +0218 : Altitude, meters, from geoid ? ellipsoid ?
     004 : speed, km/h
     013 : course, 000..359 clockwise
  950701 : fix date/time, like current d/t above
       6
  090345
       D : DOP, A=1 J=10 K=12 Q=100+ (A..J step 1, then exponential ?)
       4 : sats_in_fix, 3=2D 4=3D
       A : Geodetic datum, A=WGS84 B=Japan/Korea
8x PEASL : channel 1...8 status, 5 coded letters for each channel
           sat#, elevation, azimuth, status A..F, signal level A..Z
       D : reference oscillator status, P=Locked U=not A=? D=?
      DH : dont care
       1 : coordinate format, 0..9=deg/milliminute else deg/min/decisecond
       0 : OCR error ? O=Odd E=Even parity, of sum ? of previous bytes ?

#endif /* sony ips summary */

struct sony_ips {
	byte
		SONY[4], version[2],

		wall_yymmdd[6], wall_wday, wall_hhmmss[6],

		lat_hem, lat[7],
		lon_hem, lon[8],
		alt_sign, altitude[4], kmh[3], course[3],
		fix_yymmdd[6], fix_wday, fix_hhmmss[6],
		DOP, sats_in_fix, datum;

	struct sony_ips_channel_info {
		byte
			PRN, elevation, direction, status, siglev;
	}
		channel_info[8]; /* or 12 */
};

struct sony_ips_tail {
	byte
		oscillator_status, dont, care,
		lat_lon_format,
		parity;

	/* and \r\n makes it 110/130 bytes */
};

/* fld is one past the hemisphere letter */

static void
sony_coord(long *pmm, byte *fld, struct sony_ips_tail *q)
{
	byte hem = __tolower(fld[-1]);
	word deg, mm;

	if (hem == 'e' || hem == 'w')
		deg = sscan_03u(fld);
	else
		deg = sscan_02u(fld--); /* align rest of field */

	fld += 3;

	if (__isdigit(q->lat_lon_format))
		mm = sscan_num(fld + 0, 5);
	else
		mm = sscan_02u(fld + 0) * 1000
		   + sscan_03u(fld + 2) * 10 / 6; /* decisecond to milliminute */

	*pmm = EMUL(_DEG_AS_mm, deg) + mm;

	if (hem == 's' || hem == 'w')
		lnegate(pmm);
}

/* 0 means invalid/unknown */

static word
sony_dop(word c)
{
	static byte tab[] = { 1,2,3,4,5,6,7,8,9,10,11,13,16,21,31,51,100 };

	return ((c < 'A' || c > 'Q') ? 0 : tab[c - 'A']);
}

/* \r\n has been snipped at this point, line is ^SONY or ^SMAT */

void
sony_gps_line(byte *line)
{
	struct sony_ips *p = (void *) line;
	struct sony_ips_tail *q;
	struct sony_ips_channel_info *ci;
	int nsat, alt;
	struct tm tm;
	struct vector v;

	/* require WGS84 (datum A) and nothing else.
	 *
	 * I guess the parity algorithm will remain unknown.
	 * does not have much value anyhow, 1 poor bit.
	 * better be anal with character ranges. The line collector
	 * has already scrapped lines containing non-printables.
	 *
 	 * Again here. It seems the 'parity' _really_ is the lsbit
	 * of sum of bytes before the parity field. What on earth... */

	if (p->datum != 'A'
	 || !__isdigit(p->sats_in_fix)
	 || (p->alt_sign != '+' && p->alt_sign != '-')
	 || (p->lon_hem  != 'E' && p->lon_hem  != 'W')
	 || (p->lat_hem  != 'N' && p->lat_hem  != 'n'
	  && p->lat_hem  != 'S' && p->lat_hem  != 's'))
		return;

	switch (strlen(line)) {
	default:
		return;

	case 128:
		q = (void *) ci = &p->channel_info[12];
		break;

	case 108:
		q = (void *) ci = &p->channel_info[8];
	}

#define sony_bad_digits(p, a, b) ! is_digits( \
	__offsetof(struct sony_ips, b) - __offsetof(struct sony_ips, a), p->a)

	if ((q->parity != 'E' && q->parity != 'O')
	 || sony_bad_digits(p, wall_yymmdd, lat_hem) /* all of current_time */
	 || sony_bad_digits(p, lat,         lon_hem)
	 || sony_bad_digits(p, lon,         alt_sign)
	 || sony_bad_digits(p, altitude,    DOP))    /* + kmh course fix_time */
		return;

	/* after above checks, time is valid */

	tm.year   = sscan_02u(p->wall_yymmdd + 0);
	tm.month  = sscan_02u(p->wall_yymmdd + 2);
	tm.day    = sscan_02u(p->wall_yymmdd + 4);

	tm.hour   = sscan_02u(p->wall_hhmmss + 0);
	tm.minute = sscan_02u(p->wall_hhmmss + 2);
	tm.second = sscan_02u(p->wall_hhmmss + 4);

	nsat = 0;
	do {
		switch ((--ci)->status) {

		case 'B':   /* synchronized */
		case 'C':   /* usable for calculation */
		case 'F':   /* used in calculation */
			++nsat;
		}
	} while (p->channel_info != ci);

	gps_satellites = nsat;

	/* it would be PDOP really ? */

	HDOP = sony_dop(p->DOP);

	if (__isupper(p->lat_hem)) {

		/* means fix is current */

		sony_coord(&v.coord.lat, p->lat, q);
		sony_coord(&v.coord.lon, p->lon, q);

		v.course = sscan_03u(p->course);
		v.knots  = muldiv(sscan_03u(p->kmh), 1000, 1852); /* XXX km_to_nm */

		alt = sscan_num(p->altitude, 4);
		if (p->alt_sign == '-')
			alt = -alt;

		sted_gps_altitude(alt);

		gps_fix(&v, &tm);
	} else
		gps_nofix(&tm);
}

