
#include "standard.h"
#include "crc.h"
#include "kiss.h"

#define ld_word_moto(a) (*(word *) (a))

#define crc_SMACK(p,n,crc) crc_type_1(n,crc,0xA001,p)
#define crc_RMNC(p,n,crc)  crc_type_R(n,crc,0x8408,p)

#define iv_SMACK 0
#define iv_RMNC  ~0

/* XKISS would be, they say, a xor of the packet bytes,
 * located after the packet bytes. does it include
 * the kiss port/command byte ? */


/* port/command byte must still remain in p[0]
 * signed 'n' is just sloppiness */

int
verify_kiss_crc(byte *p, int n)
{
	word calc, rcvd;

	n -= 2;

	switch (*p & KISS_CRC) {

	/* what about when both are set ? :) */
	default:
		return (0);

	case KISS_SMACK:
		if (n < 1)
			return (-1);

		rcvd = ld_word_intel(&p[n]);
		calc = crc_SMACK(p, n, iv_SMACK);
		break;

	case KISS_RMNC:
		if (n < 1)
			return (-1);

		rcvd = ld_word_moto(&p[n]);
		calc = crc_RMNC(p, n, iv_RMNC);
		break;
	}
	if (calc != rcvd)
		return (-1);

	return (*p & KISS_CRC);
}

/* returned word is pre-swapped here so that emit(&crc, 2) works.
 * XXX portability ? whazzat ? */

word
calc_kiss_crc(byte *p, word len, byte cmd)
{
	word crc;

	if (len) {
		switch (cmd & KISS_CRC) {

		case KISS_SMACK:
			crc = crc_SMACK(p, len, crc_SMACK(&cmd, 1, iv_SMACK));

			return (ld_word_intel(&crc));       /* LSByte to go 1st */

		case KISS_RMNC:
			crc = crc_RMNC(p, len, crc_RMNC(&cmd, 1, iv_RMNC));

			return (crc);        /* MSByte to go 1st */
		}
	}
	doh("kisrc");
}

#if 0 /* in crc.s --- argument order is a femto-optimisation */

static word
crc_type_R(word n, word crc, word poly, byte *p)
{
	word x, ox;

	do {
		x = crc >> 8;
		x ^= ~0;
		x ^= *p++;

#define BIT ox = x; x >>= 1; if (ox & 1) x ^= poly;

		BIT; BIT; BIT; BIT
		BIT; BIT; BIT; BIT

		crc <<= 8;
		crc ^= x;

	} while (--n);

	return (crc);
}

#endif

