
	include regs.inc

PIT_V equ 0A4h  ; V 52h

	;------------------------------------------
	section .bankf
	align 1

	global demod_lock, rssi_avg

rssi_avg     ds.w 1   ; accessed directly from afsk.s

demod_lock   ds.w 1
txproc_lock  ds.w 1

txproc_delay ds.w 1

	;----------------------------------------------------------------------
	section .bss
	align 1

	global time, milliuptime, _gottime

milliuptime_ext ds.w 1
milliuptime     ds.w 2   ; freerunning milliseconds (approx)

time      ds.w 2   ; time as seconds since 00-01-01 00:00:00

pit_tick  ds.w 1   ; downcounts each second
_gottime  ds.b 1   ; set, when time contains valid data
          ds.b 1

_channel_utilization ds.w 2 ; dword

	;------------------------------------------
	section .inthand
	align 1

oops
	jmp stack_underflow       ; reboot

	; PIT hitting at 1024 Hz

PIT_int

	cps  # __lostack + 100    ; some margin
	blo  oops

	pshm d,e,x,y,z,k

	clrb
	tbek                           ; EXT ram

	; increment stuff and such
	; XXX if RSSI > 300 then used++  ---  try to handle collisions

	ldd  _afsk_dcd                 ; if softdcd, channel is used.
	ord  _force_dcd                ; or when tx_state > TRYING.
	cpd  # 18                      ; XXX tune for 1% with just noise
	blo  L100
	incw _channel_utilization + 2
	bne  L100
	incw _channel_utilization + 0
L100

	; binary time - not quite milliseconds, but close enough

	incw milliuptime + 2
	bne  L0
	incw milliuptime + 0  ; wrap after 4G
	bne  L0
	incw milliuptime_ext  ; extend to 48 bits
L0

	; wallclock - exact seconds

	decw pit_tick       ; 1024 times a second
	bne  L1
	ldab # 4            ; 0x400 = 1024
	stab pit_tick       ; LSByte was already zero
	incw time + 2       ; another second elapsed
	bne  L1
	incw time + 0       ; wrap in 136 years (?)
L1

	ldab # FF
	tbek             ; EXT F

	; rssi average

	ldd  rssi_avg
	asld
	asld
	asld
	subd rssi_avg   ; 7/8 old
	addd RJURR1     ; 1/8 new from RSSI adc
	lsrd
	lsrd
	lsrd
	std  rssi_avg

	; txproc timeout

	tstw txproc_delay
	beq  L3
	decw txproc_delay
L3

	orp  # 00060h          ; 2 to 3
	andp # 0FF3Fh          ; 3 to 1

	; demodulator

	decw demod_lock        ; try
	bne  L4                ; if not success
	ldab adc_buf_remptr + 1
	cmpb adc_buf_insptr + 1
	beq  L4

	jsr  afsk_rx_dsp       ; all regs free for use
	ldab # FF
	tbek
L4
	incw demod_lock        ; undo

	; txproc

	tstw txproc_delay
	bne  L6                ; not yet
	decw txproc_lock       ; try
	bne  L5                ; if not success

	clrb
	tbek
	tbxk
	tbyk                   ; banks as "C" likes them

	ldx  _txproc           ; variable in bank 0, pointing to routines in bank 3
	jsr  text_start,x      ; all regs free for use
	tde

	ldab # FF
	tbek
	ste  txproc_delay
L5
	incw txproc_lock       ; undo
L6

	pulm d,e,x,y,z,k
	rti


	;------------------------------------------
	section .text
	align 1

	;---------------------------------------
	global _init_PIT

_init_PIT

	pshm k,ccr
	orp  # 000E0h

	clrb
	tbek               ; external RAM

	ldd  # PIT_int
	std  PIT_V         ; vector 52h at address A4h

	ldd  # 1024
	std  pit_tick

	ldab # FF
	tbek               ; SFRs and internal ram

	ldd  # 0252h       ; IPL 2 IV 52h
	std  PICR
	ldd  # 8           ; 1024 Hz
	std  PITR

	pulm k,ccr
	rts

	;---------------------------------------
	global _channel_usage_percent

_channel_usage_percent

	xgex                ; 0,x == mark

	lded _channel_utilization
	pshm e,d

	clre
	clrd
	sted _channel_utilization

	lded milliuptime
	pshm e,d

	tsy                 ; 0,y == now, 4,y == n

	lde  0,y
	ldd  2,y
	subd 2,x
	sbce 0,x            ; elapsed in E:D
L102
	lsrw 4,y
	rorw 6,y            ; n >>= 1
	lsre
	rord                ; elapsed >>= 1
	tste
	bne  L102           ; while elapsed too wide

	lde  0,y
	ste  0,x
	lde  2,y
	ste  2,x            ; mark = now

	lde  6,y            ; n in E. elapsed in D, both scaled

	ais  #+8            ; stack purged

	ldx  # 100
	tstd
	beq  L103           ; huh? elapsed == 0
	xgdx                ; elapsed in x, 100 in D
	emul                ; E:D = 100 * n
	ediv                ; x = 100 * n / elapsed
L103
	xgdx                ; in D
	rts

	;---------------------------------------
	global _now, _future_msec, _future_sec
	global _time_past, _msec_since
	global _delay_ms

_now              ; void now(stamp *t)
	xgex
	lded milliuptime
	ste  0,x
	std  2,x
	rts

_future_sec       ; void future_sec(stamp *t_E, word sec_D)
	xgex          ; pointer to stamp
	lde  # 1000   ; a second
	emul          ; ED is msec
	ste  0,x      ; msec MSWord
	bra  L30

_future_msec      ; void future_msec(stamp *t, word msec)
	xgex
	clrw 0,x      ; msec MSWord
L30
	std  2,x      ; msec LSWord
	lded milliuptime
	addd 2,x      ; msec
	std  2,x
	adce 0,x      ; ... MSWord
	ste  0,x
	rts

_msec_since        ; word msec_since(stamp *then)
	xgex
	lded milliuptime
	subd 2,x
	sbce 0,x       ; spent = now - then
	beq  L10
	clre
	clrd
	comd
L10
	rts

_time_past         ; bool time_past(stamp) XXX rethink this !
	pshm d,e
	tsx            ; E,D at x
	lded milliuptime
	subd 2,x
	sbce 0,x       ; ED negative if not yet passed that time
	come           ; positive  ...
	clrd
	asle           ; carry if passed
	rolb           ; FALSE or TRUE in D
	ais  # 4       ; purge argument dword
	rts

_delay_ms           ; not for delays over 30 seconds
	adde milliuptime + 2   ; target time
L7
	cpe  milliuptime + 2   ; XXX could/should wai ?
	bpl  L7
	rts

	;---------------------------------------
	global _read_rssi_avg

_read_rssi_avg
	pshm k
	ldab # FF
	tbek
	ldd  rssi_avg
	pulm k
	rts

	;---------------------------------------
	global _unleash_txproc, _hold_txproc, _okay_txproc

_unleash_txproc
	pshm k
	ldab # FF
	tbek
	ldd  # 1
	std  txproc_lock
	pulm k
	rts

_hold_txproc
	pshm k
	ldab # FF
	tbek
	decw txproc_lock
	pulm k
	rts

_okay_txproc
	pshm k
	ldab # FF
	tbek
	incw txproc_lock
	pulm k
	rts

	;----------------------------------------------------------------------
	global _time, _uptime, _bolt, _settime, _wallclock_reset

_wallclock_reset         ; void (struct tm *);

	jsr  _mktime_regular ; pass arg in E and return time in ED for ...

_settime                 ; void (time_t);

	ldx  # 1024          ; "next second in a sec"
	stx  pit_tick        ; downcounter rewind

	sted time

	ldab # 1
	stab _gottime
	rts

_time
	xgex
	lded time
	ste  0,x
	std  2,x
	rts

_uptime
_bolt
	lded milliuptime_ext + 1 ; quarter seconds
	lsre
	rord
	lsre
	rord
	rts                    ; seconds (wrap after 30+ years)

	;---------------------------------------
	end

