
#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "stat.h"
#include "ioctl.h"
#include "socket.h"
#include "socketvar.h"
#include "protosw.h"
#include "domain.h"
#include "mbuf.h"

#include "net/if.h"
#include "net/route.h"
#include "net/af.h"

#include "ax25/ax.h"
#define AXSTATES
#include "ax25/ax_pcb.h"
#include "ax25/ax_var.h"

static struct ax_addr zero_axaddr;
static struct axpcb *last_axpcb = 0;
static struct axpcb *last_axpcbeacon = 0;

ax_pcbprint(axp)
	struct axpcb *axp;
{
	if (axp->axp_ifp)
		printf("%s%d ", axp->axp_ifp->if_name, axp->axp_ifp->if_unit);
	else
		printf("NULLIFP ");

	printf("%s %s ", ax_sprintf(&axp->axp_faddr), ax_sprintf(&axp->axp_laddr));

	printf("%d/%d/%d %d ",
		axp->axp_T1, axp->axp_T2, axp->axp_T3, axp->axp_N2);

	printf("s%d r%d u%d o%d ",
		(unsigned) axp->axp_vs, (unsigned) axp->axp_vr,
		(unsigned) axp->axp_una, axp->axp_off);

	if (axp->axp_pollsent)
		printf("poll ");
	if (axp->axp_rejsent)
		printf("rej ");
	if (axp->axp_remotebusy)
		printf("rbusy ");
	if (axp->axp_localbusy)
		printf("lbusy ");

	printf("lrtt %d srtt %d ", axp->axp_lastrtt, axp->axp_srtt);

	printf("%x %s\n", axp->axp_flags, axstates[axp->axp_state]);
}

struct axpcb *
ax_pcblookup(head, ifp, faddr, laddr, matchwild)
	struct axpcb *head;
	struct ifnet *ifp;
	struct ax_addr *faddr, *laddr;
{
	struct axpcb *maybe = 0;
	register struct axpcb *axp = last_axpcb;

	if (axp && ifp == axp->axp_ifp
			&& ax_addreq(faddr, &axp->axp_faddr) > 1
			&& ax_addreq(laddr, &axp->axp_laddr) > 1)
		return (axp);

	for (axp = head->axp_next; axp != head; axp = axp->axp_next) {
		if (ifp != axp->axp_ifp)
			continue;
		if (axp->axp_laddr.ax_call[0] == 0)
			continue;
		if (ax_addreq(laddr, &axp->axp_laddr) <= 1)
			continue;
		if (axp->axp_faddr.ax_call[0] == 0) {
			maybe = axp;
			continue;
		}
		if (ax_addreq(faddr, &axp->axp_faddr) > 1) {
			last_axpcb = axp;
			return (axp);
		}
	}
	return (matchwild ? maybe : NULL);
}

struct axpcb *
ax_pcblookupbeacon(ifp, dst)
	struct ifnet *ifp;
	struct ax_addr *dst;
{
	register struct axpcb *axp = last_axpcbeacon;

	if (axp && ifp == axp->axp_ifp
			&& ax_addreq(dst, &axp->axp_faddr) > 1)
		return (axp);

	for (axp = axbeacons.axp_next; axp != &axbeacons; axp = axp->axp_next) {
		if (ifp != axp->axp_ifp)
			continue;
		if (ax_addreq(dst, &axp->axp_faddr) > 1) {
			last_axpcbeacon = axp;
			return (axp);
		}
	}
	return (NULL);
}

ax_pcballoc(so, head)
	struct socket *so;
	struct axpcb *head;
{
	struct mbuf *m;
	struct axpcb *axp;

	m = m_getclr(M_DONTWAIT, MT_PCB);
	if (m == 0)
		return (ENOBUFS);
	axp = mtod(m, struct axpcb *);
	axp->axp_head = head;
	axp->axp_socket = so;
	insque(axp, head);
	so->so_pcb = (caddr_t) axp;
	return 0;
}

ax_pcbconnect(axp, nam)
	struct axpcb *axp;
	struct mbuf *nam;
{
	struct sockaddr_ax *sax = mtod(nam, struct sockaddr_ax *);
	struct ax_ifaddr *ia;
	int ssid;

	if (ax_debug > 4)
		{ printf("pcbconnect:"); ax_pcbprint(axp); }

	if (nam->m_len != sizeof (*sax))
		return (EINVAL);
	if (sax->sax_family != AF_AX25)
		return (EAFNOSUPPORT);

	if (axp->axp_ifp == NULL) {

		if ((ia = ax_ifaddr) == NULL)
			return (EADDRNOTAVAIL);

		axp->axp_laddr = ia->aa_addr.sax_addr;
		ssid = 0;
		for (;;) {
			axp->axp_laddr.ax_ssid = (ssid << 1);
			if (ax_pcblookup(axp->axp_head, ia->aa_ifp,
					&sax->sax_addr, &axp->axp_laddr, 0) == NULL)
				break;
			if (++ssid > 15)
				return (EADDRINUSE);
		}
		axp->axp_ifp = ia->aa_ifp;
	} else {
		if (ax_pcblookup(axp->axp_head, axp->axp_ifp,
				&sax->sax_addr, &axp->axp_laddr, 0))
			return (EADDRINUSE);
	}
	axp->axp_faddr = sax->sax_addr;

	return (0);
}

ax_pcbbind(axp, nam)
	struct axpcb *axp;
	struct mbuf *nam;
{
	if (ax_debug > 4)
		{ printf("pcbbind:"); ax_pcbprint(axp); }

	return ax_bindlocal(nam, axp->axp_head, &axp->axp_ifp, &axp->axp_laddr);
}

ax_setsockaddr(axp, nam)
	struct axpcb *axp;
	struct mbuf *nam;
{
	struct sockaddr_ax *sax = mtod(nam, struct sockaddr_ax *);

	nam->m_len = sizeof (*sax);
	bzero(sax, sizeof(*sax));
	sax->sax_family = AF_AX25;
	sax->sax_len = sizeof(*sax);
	sax->sax_addr = axp->axp_laddr;
	ax_ifsprintf(sax->sax_ifname, axp->axp_ifp);
}

ax_setpeeraddr(axp, nam)
	struct axpcb *axp;
	struct mbuf *nam;
{
	struct sockaddr_ax *sax = mtod(nam, struct sockaddr_ax *);

	nam->m_len = sizeof (*sax);
	bzero(sax, sizeof(*sax));
	sax->sax_family = AF_AX25;
	sax->sax_len = sizeof(*sax);
	sax->sax_addr = axp->axp_faddr;
	ax_ifsprintf(sax->sax_ifname, axp->axp_ifp);
}

ax_pcbdisconnect(axp)
	struct axpcb *axp;
{
	if (ax_debug > 4)
		{ printf("disconnect:"); ax_pcbprint(axp); }

	axp->axp_faddr.ax_call[0] = 0;

	if (axp->axp_socket->so_state & SS_NOFDREF)
		ax_pcbdetach(axp);
}

ax_pcbdetach(axp)
	struct axpcb *axp;
{
	struct socket *so = axp->axp_socket;

	if (ax_debug > 4)
		{ printf("detach:"); ax_pcbprint(axp); }

	if (axp == last_axpcb)
		last_axpcb = NULL;
	if (axp == last_axpcbeacon)
		last_axpcbeacon = NULL;

	so->so_pcb = 0;
	sofree(so);
	remque(axp);
#ifdef notyet
	if (axp->axp_route.ro_rt)
		rtfree(axp->axp_route.ro_rt);
#endif
	(void) m_free(dtom(axp));
}

ax_ifsprintf(s, ifp)
	char *s;
	struct ifnet *ifp;
{
	if (ifp) {
		char *n = ifp->if_name;

		while (*s = *n++)
			s++;
		if (ifp->if_unit >= 10)
			*s++ = '0' + (ifp->if_unit / 10);
		*s++ = '0' + (ifp->if_unit % 10);
	}
	*s = 0;
}

/*
 * Bind to a local address.
 *
 * User can specify the interface,
 * the callsign, and the ssid.
 */

ax_bindlocal(nam, head, ifpp, addr)
	struct mbuf *nam;
	struct axpcb *head;
	struct ifnet **ifpp;
	struct ax_addr *addr;
{
	struct ax_ifaddr *ifa;

	if (nam == NULL) {
		/*
		 * Nothing specified (autobinding on connect()).
		 * Use 1st configured interface.
		 */
		struct ax_addr local;
		int ssid;

		if ((ifa = ax_ifaddr) == NULL)
			return (EADDRNOTAVAIL);

		local = ifa->aa_addr.sax_addr;
		ssid = 0;
		for (;;) {
			local.ax_ssid = (ssid << 1);
			if (ax_pcblookup(head, ifa->aa_ifp,
					&zero_axaddr, &local, 1) == NULL)
				break;
			if (++ssid > 15)
				return (EADDRINUSE);
		}
		*addr = local;
	} else {
		struct sockaddr_ax *sax;

		if (nam->m_len != sizeof (*sax))
			return (EINVAL);
		sax = mtod(nam, struct sockaddr_ax *);
		if (sax->sax_family != AF_AX25)
			return (EAFNOSUPPORT);

		/*
		 * 1) Interface specified.
		 * 2) Local callsign specified.
		 * 3) Neither.
		 */
		if (sax->sax_ifname[0])
			ifa = ax_ifawithifname(sax->sax_ifname);
		else
		if (sax->sax_addr.ax_call[0])
			ifa = ax_ifawithaddr(&sax->sax_addr);
		else
			ifa = ax_ifawithssid(&sax->sax_addr);

		if (ifa == NULL)
			return (EADDRNOTAVAIL);

		bcopy(ifa->aa_addr.sax_addr.ax_call, sax->sax_addr.ax_call, ALEN);

		if (ax_pcblookup(head, ifa->aa_ifp,
				&zero_axaddr, &sax->sax_addr, 1))
			return (EADDRINUSE);

		*addr = sax->sax_addr;
	}
	*ifpp = ifa->aa_ifp;
	return (0);
}

